/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record;

import java.util.HashMap;
import java.util.Map;

import javax.script.ScriptContext;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import javax.script.SimpleScriptContext;

/**
 * Class Expr supports evaluation of script expressions at runtime.
 */
public class Expr {
	private static ScriptEngineManager sem = new ScriptEngineManager();
	private static Map<String, ScriptEngine> eng = new HashMap<String, ScriptEngine>();
	/**
	 * Evaluate script expression.
	 * @param expr expresssion in format "language:expression"
	 * @param rec record (object available as rec in expression)
	 * @return int value of expression
	 * @throws RecordException if error
	 */
	public static synchronized int eval(String expr, Object rec) throws RecordException {
		int ix = expr.indexOf(':');
		String lang = expr.substring(0,  ix);
		String expr2 = expr.substring(ix + 1);
		if(!eng.containsKey(lang)) {
			eng.put(lang, sem.getEngineByName(lang));
		}
		ScriptEngine se = eng.get(lang);
        ScriptContext ctx = new SimpleScriptContext();
	    ctx.setAttribute("rec", rec, ScriptContext.ENGINE_SCOPE);
        Object res;
		try {
			res = se.eval(expr2, ctx);
		} catch (ScriptException e) {
			throw new RecordException("Exception evaluating expression '" + expr + "': " + e.getMessage());
		}
        if(res instanceof Integer) {
        	return (int)(Integer)res;
        } else if(res instanceof Long) {
        	return (int)(long)(Long)res;
        } else if(res instanceof Double) {
        	return (int)(double)(Double)res;
        } else if(res instanceof String) {
        	return Integer.parseInt((String)res);
        } else {
        	throw new RecordException("Unsupported expression return type: " + res.getClass().getName());
        }
	}
}
