/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record.test;

import static org.junit.Assert.*;

import org.junit.Test;

import dk.vajhoej.record.Expr;
import dk.vajhoej.record.RecordException;

public class TestExpr {
	public static class DemoRec {
		private int v;
		public DemoRec(int v) {
			this.v = v;
		}
		public int getV() {
			return v;
		}
		public void setV(int v) {
			this.v = v;
		}
	}
    @Test
    public void testBasic() {
    	try {
    		DemoRec rec = new DemoRec(123);
			int v = Expr.eval("js:rec.v", rec);
			assertEquals("v", 123, v);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
    @Test
    public void testExpr() {
    	try {
    		DemoRec rec = new DemoRec(123);
			int rem = Expr.eval("js:rec.v % 2", rec);
			assertEquals("rem", 1, rem);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
}
