package dk.vajhoej.record.test;

import static org.junit.Assert.*;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;
import java.util.TimeZone;

import org.junit.Test;

import dk.vajhoej.record.CodeGen;
import dk.vajhoej.record.CompositFastStructWriter;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.FastStructWriter;
import dk.vajhoej.record.StructInfoCache;
import dk.vajhoej.record.StructWriter;

public class TestFastStructWriter {
	private final static int N = 100000;
	@Test
	public void testWriteSimple() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
				     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			FastStructWriter<Data> fsw = CodeGen.genWriter(Data.class, 100*N); // needed by testPerf later !!
			Data o = new Data();
			o.setIv(258);
			o.setXv(123.456);
			o.setSv("ABC     ");
			fsw.write(o);
			byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
			for(int i = 0; i < res.length; i++) {
				assertEquals("byte " + i, b[i], res[i]);
			}
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testWriteMulti() {
		byte[] b = { 0x01, 0x00, 0x00, 0x00,
			         0x01, 0x00, 0x00, 0x00,
		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
		             0x03, 0x00, 0x00, 0x00,
			         0x02, 0x00, 0x00, 0x00,
			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			Map<Class<?>,FastStructWriter<?>> map = new HashMap<Class<?>,FastStructWriter<?>>();
			map.put(SubDataOne.class, CodeGen.genWriter(SubDataOne.class));
			map.put(SubDataTwo.class, CodeGen.genWriter(SubDataTwo.class));
			FastStructWriter<Object> fsw = new CompositFastStructWriter(map);
			SubDataOne o1 = new SubDataOne();
			o1.setId(1);
			o1.setTyp(1);
			o1.setX(123.456);
			SuperData o1m = o1;
			fsw.write(o1m);
			SubDataTwo o2 = new SubDataTwo();
			o2.setId(3);
			o2.setTyp(2);
			o2.setS("ABC         ");
			SuperData o2m = o2;
			fsw.write(o2m);
			byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
			for(int i = 0; i < res.length; i++) {
				assertEquals("byte " + i, b[i], res[i]);
			}
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testWriteMultiAlt() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
        			 0x01, 0x00, 0x00, 0x00,
		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
		             0x03, 0x00, 0x00, 0x00,
		             0x02, 0x00, 0x00, 0x00,
		             0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
        try {
            FastStructWriter<SubDataOne> fsw1 = CodeGen.genWriter(SubDataOne.class);
            SubDataOne o1 = new SubDataOne();
            o1.setId(1);
            o1.setTyp(1);
            o1.setX(123.456);
            fsw1.write(o1);
            FastStructWriter<SubDataTwo> fsw2 = CodeGen.genWriter(SubDataTwo.class);
            fsw2.workOn(fsw1.getByteBuffer());
            SubDataTwo o2 = new SubDataTwo();
            o2.setId(3);
            o2.setTyp(2);
            o2.setS("ABC         ");
            fsw2.write(o2);
            byte[] res = fsw2.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteMultiPad() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
                0x01, 0x00, 0x00, 0x00,
                0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40, 0x00, 0x00, 0x00, 0x00,
                0x03, 0x00, 0x00, 0x00,
                0x02, 0x00, 0x00, 0x00,
                0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
        try {
			Map<Class<?>,FastStructWriter<?>> map = new HashMap<Class<?>,FastStructWriter<?>>();
			map.put(SubDataOnePad.class, CodeGen.genWriter(SubDataOnePad.class));
			map.put(SubDataTwoPad.class, CodeGen.genWriter(SubDataTwoPad.class));
			FastStructWriter<Object> fsw = new CompositFastStructWriter(map);
            SubDataOnePad o1 = new SubDataOnePad();
            o1.setId(1);
            o1.setTyp(1);
            o1.setX(123.456);
            SuperDataPad o1m = o1;
            fsw.write(o1m);
            SubDataTwoPad o2 = new SubDataTwoPad();
            o2.setId(3);
            o2.setTyp(2);
            o2.setS("ABC         ");
            SuperDataPad o2m = o2;
            fsw.write(o2m);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testWritePerf() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
    			     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
	    		     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			long t1 = System.currentTimeMillis();
			FastStructWriter<Data> fsw = CodeGen.genWriter(Data.class, 100*N);
			for(int i = 0; i < N; i++) {
				Data o = new Data();
				o.setIv(258);
				o.setXv(123.456);
				o.setSv("ABC     ");
				fsw.write(o);
			}
			byte[] res = fsw.getBytes();
			long t2 = System.currentTimeMillis();
			assertEquals("length", N*b.length, res.length);
			assertTrue("time", t2-t1 < 0.1*N);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testWriteRelPerf() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
   		             0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			long t1 = System.currentTimeMillis();
			StructWriter sw = new StructWriter(100*N);
			for(int i = 0; i < N; i++) {
				Data o = new Data();
				o.setIv(258);
				o.setXv(123.456);
				o.setSv("ABC     ");
				sw.write(o);
			}
			byte[] res = sw.getBytes();
			long t2 = System.currentTimeMillis();
			assertEquals("length", N*b.length, res.length);
			long t1f = System.currentTimeMillis();
			FastStructWriter<Data> fsw = CodeGen.genWriter(Data.class, 100*N);
			for(int i = 0; i < N; i++) {
				Data o = new Data();
				o.setIv(258);
				o.setXv(123.456);
				o.setSv("ABC     ");
				fsw.write(o);
			}
			byte[] resf = fsw.getBytes();
			long t2f = System.currentTimeMillis();
			assertEquals("fast length", N*b.length, resf.length);
			assertTrue("relative time", (t2f - t1f) * 1.0 / (t2 - t1) < 0.2);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testWriteCache() {
		try {
			StructInfoCache.getInstance().reset();
			FastStructWriter<Data> fsw = CodeGen.genWriter(Data.class, 100*N); // needed by testPerf later !!
			Data o = new Data();
			o.setIv(258);
			o.setXv(123.456);
			o.setSv("ABC     ");
			fsw.write(o);
			byte[] res = fsw.getBytes();
			assertEquals("length", 20, res.length);
			assertTrue("hit rate", StructInfoCache.getInstance().getHitRate() < 0.001);
			for(int i = 0; i < 9; i++) {
				FastStructWriter<Data> fsw2 = CodeGen.genWriter(Data.class);
				Data o2 = new Data();
				o2.setIv(258);
				o2.setXv(123.456);
				o2.setSv("ABC     ");
				fsw2.write(o2);
				byte[] res2 = fsw2.getBytes();
				assertEquals("length", 20, res2.length);
			}
			assertTrue("hit rate", StructInfoCache.getInstance().getHitRate() < 0.001);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testWriteBoolean() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
        try {
            FastStructWriter<BooleanData> fsw = CodeGen.genWriter(BooleanData.class);
            BooleanData o = new BooleanData();
            o.setB1(true);
            o.setB2(false);
            o.setB3(true);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testWriteBits() {
		byte[] b = { 0x01, 0x23, 0x04, (byte)0x80, 0x02 };
		try {
			FastStructWriter<BitData> fsw = CodeGen.genWriter(BitData.class);
			BitData o = new BitData();
			o.setI1(1);
			o.setN1(2);
			o.setN2(3);
			o.setI2(4);
			o.setZ1(1);
			o.setZ2(2);
			fsw.write(o);
			byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
			for(int i = 0; i < res.length; i++) {
				assertEquals("byte " + i, b[i], res[i]);
			}
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testWriteTime() {
		byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			         0x10, 0x27, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			         0x01, 0x00, 0x00, 0x00,
			         0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
		try {
			FastStructWriter<TimeData> fsw = CodeGen.genWriter(TimeData.class);
			TimeData o = new TimeData();
			Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
			cal.set(Calendar.YEAR, 1970);
			cal.set(Calendar.MONTH, Calendar.JANUARY);
			cal.set(Calendar.DAY_OF_MONTH, 1);
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 1);
			o.setT1(cal.getTime());
			cal.set(Calendar.YEAR, 1858);
			cal.set(Calendar.MONTH, Calendar.NOVEMBER);
			cal.set(Calendar.DAY_OF_MONTH, 17);
			cal.set(Calendar.HOUR_OF_DAY, -TimeZone.getDefault().getOffset(System.currentTimeMillis()) / 3600000);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 1);
			o.setT2(cal.getTime());
			cal.set(Calendar.YEAR, 1970);
			cal.set(Calendar.MONTH, Calendar.JANUARY);
			cal.set(Calendar.DAY_OF_MONTH, 1);
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 1);
			cal.set(Calendar.MILLISECOND, 0);
			o.setT3(cal.getTime());
			cal.set(Calendar.YEAR, 1970);
			cal.set(Calendar.MONTH, Calendar.JANUARY);
			cal.set(Calendar.DAY_OF_MONTH, 1);
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 1);
			cal.set(Calendar.MILLISECOND, 0);
			o.setT4(cal.getTime());
			fsw.write(o);
			byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
			for(int i = 0; i < res.length; i++) {
				assertEquals("byte " + i, b[i], res[i]);
			}
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testWriteBCD() {
        byte[] b = { 0x12, 0x34, 0x56, 0x7D,
                    (byte)0xF1, (byte)0xF2, (byte)0xF7, (byte)0xF9, (byte)0xF5, (byte)0xC0  };
        try {
            FastStructWriter<BCDData> fsw = CodeGen.genWriter(BCDData.class);
            BCDData o = new BCDData();
            o.setV1(new BigDecimal("-12345.67"));
            o.setV2(new BigDecimal("1279.50"));
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteVaxFloat() {
        byte[] b = { 0x45, 0x42, (byte)0xA4, 0x70,
                     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE };
        try {
            FastStructWriter<VaxFloatData> fsw = CodeGen.genWriter(VaxFloatData.class);
            VaxFloatData o = new VaxFloatData();
            o.setV1(12.34f);
            o.setV2(123.456);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteVaxFloatArray() {
        byte[] b = { (byte)0x99, 0x40, (byte)0x9A, (byte)0x99,
         		     0x45, 0x42, (byte)0xA4, 0x70,
   		             (byte)0xF6, 0x43, 0x79, (byte)0xE9,
   		             0x00, 0x00, 0x00, 0x00,
           		     0x00, (byte)0xC0, 0x00, 0x00,
   		             0x20, (byte)0xC1, 0x00, 0x00,
        		     (byte)0xDB, 0x2F, (byte)0xFF, (byte)0xE6,
        		     0x15, 0x51, (byte)0xF9, 0x02,
        		     0x13, 0x40, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33,
        		     0x48, 0x40, 0x14, (byte)0xAE, (byte)0xE1, 0x7A, (byte)0xAE, 0x47,
        		     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE,
        		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x24, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     (byte)0xFB, 0x3D, (byte)0xDF, 0x7C, (byte)0xD7, (byte)0xD9, (byte)0xBB, (byte)0xBD,
         		     0x22, 0x42, 0x5F, (byte)0xA0, 0x00, 0x20, 0x00, 0x00,
         		     (byte)0x99, 0x40, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x98, (byte)0x99,
        		     0x45, 0x42, (byte)0xA3, 0x70, 0x0A, (byte)0xD7, 0x70, 0x3D,
        		     (byte)0xF6, 0x43, 0x78, (byte)0xE9, (byte)0xFD, (byte)0xD4, (byte)0xB8, (byte)0xF3,
        		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x20, (byte)0xC1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     (byte)0xDB, 0x2F, (byte)0xFE, (byte)0xE6, (byte)0xBD, (byte)0xCE, (byte)0xD8, (byte)0xED,
   		             0x15, 0x51, (byte)0xF9, 0x02, 0x00, 0x00, 0x00, 0x00 };
        float[] x1 = { 1.2f, 12.34f, 123.456f, 0.0f, -0.5f, -2.5f, 0.0000000001f, 10000000000.0f };
        double[] x2 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
        double[] x3 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
        try {
            FastStructWriter<VaxFloatArrayData> fsw = CodeGen.genWriter(VaxFloatArrayData.class);
            VaxFloatArrayData o = new VaxFloatArrayData();
            o.setV1(x1);
            o.setV2(x2);
            o.setV3(x3);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteUnsigned() {
        byte[] b = { (byte)0xFF,
                     (byte)0xFF, (byte)0xFF,
                     (byte)0xFF, (byte)0xFF,(byte)0xFF, (byte)0xFF };
        try {
            FastStructWriter<UnsignedData> fsw = CodeGen.genWriter(UnsignedData.class);
            UnsignedData o = new UnsignedData();
            o.setUi1v((short)255);
            o.setUi2v(65535);
            o.setUi4v(4294967295L);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteArray() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x03, 0x01, 0x00, 0x00,
                     0x04, 0x01, 0x00, 0x00,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
        try {
            FastStructWriter<ArrayData> fsw = CodeGen.genWriter(ArrayData.class);
            ArrayData o = new ArrayData();
            for(int ix = 0; ix < 3; ix++) {
                o.setIv(ix, 258 + ix);
                o.setXv(ix, 123.456);
            }
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteCountArray() {
        byte[] b = { 0x04,
        		     0x02, 0x01, 0x00, 0x00,
	                 0x03, 0x01, 0x00, 0x00,
	                 0x04, 0x01, 0x00, 0x00,
	                 0x05, 0x01, 0x00, 0x00,
	                 0x04, 0x00,
	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
        try {
            FastStructWriter<CountArrayData> fsw = CodeGen.genWriter(CountArrayData.class);
            CountArrayData o = new CountArrayData();
            for(int ix = 0; ix < 4; ix++) {
                o.setIv(ix, 258 + ix);
                o.setXv(ix, 123.456);
            }
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testWriteStructField() {
		try {
			FastStructWriter<MainData> fsw = CodeGen.genWriter(MainData.class);
			MainData o = new MainData();
			o.setI1(1);
			o.setI2(2);
			o.s = new FieldData();
			o.s.setI3(3);
			o.s.setI4(4);
			o.setI5(5);
			fsw.write(o);
			fail("Struct fields not supported");
		} catch (RecordException e) {
		}
	}
    @Test
    public void testWriteStructAlignment0() {
        byte[] b = { 0x01, 0x02, 0x00, 0x03,
                     0x00, 0x00, 0x00, 0x04,
                     0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00 };
        try {
            FastStructWriter<AlignData0> fsw = CodeGen.genWriter(AlignData0.class);
            AlignData0 o = new AlignData0();
            o.setI1((byte) 1);
            o.setI2((short) 2);
            o.setI4(3);
            o.setI8(4);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteStructAlignment1() {
        byte[] b = { 0x01, 0x00, 0x02, 0x00,
                     0x03, 0x00, 0x00, 0x00,
                     0x04, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00 };
        try {
            FastStructWriter<AlignData1> fsw = CodeGen.genWriter(AlignData1.class);
            AlignData1 o = new AlignData1();
            o.setI1((byte) 1);
            o.setI2((short) 2);
            o.setI4(3);
            o.setI8(4);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteStructAlignment2() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x02, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x03, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x04, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00 };
        try {
            FastStructWriter<AlignData2> fsw = CodeGen.genWriter(AlignData2.class);
            AlignData2 o = new AlignData2();
            o.setI1((byte) 1);
            o.setI2((short) 2);
            o.setI4(3);
            o.setI8(4);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testWriteError1() {
		try {
			FastStructWriter<ErrorData1> fsw = CodeGen.genWriter(ErrorData1.class);
			ErrorData1 o = new ErrorData1();
			o.setSv("ABC");
			fsw.write(o);
			fail("Missing exception");
		} catch (RecordException e) {
		}
	}
	@Test
	public void testWriteError2() {
		try {
			FastStructWriter<ErrorData2> fsw = CodeGen.genWriter(ErrorData2.class);
			ErrorData2 o = new ErrorData2();
			o.setIv(123);
			fsw.write(o);
			fail("Missing exception");
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		} catch (IllegalArgumentException e) {
		}
	}
	@Test
	public void testWriteError3() {
		try {
			FastStructWriter<ErrorData3> fsw = CodeGen.genWriter(ErrorData3.class);
			ErrorData3 o = new ErrorData3();
			o.setIv1(123);
			o.setIv2(456);
			fsw.write(o);
			fail("Missing exception");
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		} catch (IllegalArgumentException e) {
		}
	}
	@Test
    public void testWriteStrings() {
        byte[] b = { 0x41, 0x42, 0x43, 0x44,
                     0x02, 0x00, 0x41, 0x42,
                     0x02, 0x00, 0x41, 0x42, 0x00, 0x00,
                     0x41, 0x42, 0x00, 0x00,
                     0x02, 0x41, 0x42,
                     0x41, 0x42, 0x20, 0x020 };
        try {
            FastStructWriter<StringData> fsw = CodeGen.genWriter(StringData.class);
            StringData o = new StringData();
            o.setS1("ABCD");
            o.setS2("AB");
            o.setS3("AB");
            o.setS4("AB");
            o.setS5("AB");
            o.setS6("AB");
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
    public void testWriteStringsNull() {
        byte[] b = { 0x20, 0x20, 0x20, 0x20,
                     0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x00, 
                     0x20, 0x20, 0x20, 0x20 };
        try {
            FastStructWriter<StringData> fsw = CodeGen.genWriter(StringData.class);
            StringData o = new StringData();
            o.setS1(null);
            o.setS2(null);
            o.setS3(null);
            o.setS4(null);
            o.setS5(null);
            o.setS6(null);;
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    public void testWriteBigEndian() {
        byte[] b = { 0x00, 0x00, 0x01, 0x02 };
        try {
            FastStructWriter<BigEndianData> fsw = CodeGen.genWriter(BigEndianData.class);
            BigEndianData o = new BigEndianData();
            o.setIv(258);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteMixedEndian() {
        byte[] b = { 0x00, 0x01, 0x02, 0x00, 0x00, 0x03, 0x00, 0x04 };
        try {
            FastStructWriter<MixedEndianData> fsw = CodeGen.genWriter(MixedEndianData.class);
            MixedEndianData o = new MixedEndianData();
            o.setIv1(1);
            o.setIv2(2);
            o.setIv3(3);
            o.setIv4(4);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteEndPad1() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x01 };
        try {
            FastStructWriter<EndPadData1> fsw = CodeGen.genWriter(EndPadData1.class);
            EndPadData1 o = new EndPadData1();
            o.setIv(258);
            o.setBv((byte)1);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteEndPad2() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00 };
        try {
            FastStructWriter<EndPadData2> fsw = CodeGen.genWriter(EndPadData2.class);
            EndPadData2 o = new EndPadData2();
            o.setIv(258);
            o.setBv((byte)1);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testWriteEndPad3() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        try {
            FastStructWriter<EndPadData3> fsw = CodeGen.genWriter(EndPadData3.class);
            EndPadData3 o = new EndPadData3();
            o.setIv(258);
            o.setXv(0.0);
            o.setBv((byte)1);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void TestWriteGeneralInt() {
        byte[] b = { 0x03, 0x02, 0x01, 0x05, 0x04, 0x03, 0x02, 0x01 };
        try {
            FastStructWriter<GeneralIntData> fsw = CodeGen.genWriter(GeneralIntData.class);
            GeneralIntData o = new GeneralIntData();
            o.setV1(0x0000000000010203L);
            o.setV2(0x0000000102030405L);
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void TestWriteRemainingString()
    {
        byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x41, 0x42, 0x43 };
        try {
            FastStructWriter<RemainingStringData> fsw = CodeGen.genWriter(RemainingStringData.class);
            RemainingStringData o = new RemainingStringData();
            o.setIv(258);
            o.setSv("ABC");
            fsw.write(o);
            byte[] res = fsw.getBytes();
            assertEquals("length", b.length, res.length);
            for(int i = 0; i < res.length; i++) {
                assertEquals("byte " + i, b[i], res[i]);
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
}
